#include "StdAfx.h"
#include "UartDrv.h"

#ifdef _DEBUG
static void ErrorDlg (char* msg)
{
  AfxMessageBox (msg, MB_ICONEXCLAMATION | MB_OK, NULL);
//  printf ("%s\n", msg);
}
#else
#define ErrorDlg(msg)
#endif

#define RX_BUFFER_SIZE          2048
#define TX_BUFFER_SIZE          1024

#define READ_PORTION_MAX_SIZE    512

#define THREAD_EXIT_TIMEOUT     INFINITE

DWORD RxThreadFxn (LPVOID lParam);
DWORD TxThreadFxn (LPVOID lParam);

CUartDrv::CUartDrv () :
          QuitEvent (FALSE, TRUE), RxBuffer (RX_BUFFER_SIZE), TxBuffer(TX_BUFFER_SIZE)
{
  IsOpen = 0;
  CommHandle = INVALID_HANDLE_VALUE;
  pRxThread = NULL;
  pTxThread = NULL;
}

CUartDrv::~CUartDrv()
{
  ClosePort();
}

int CUartDrv::IsPortOpen()
{
  return IsOpen;
}

int CUartDrv::PutChar (unsigned char Char)
{
  return TxBuffer.PutChar (Char);
}

int CUartDrv::GetChar (unsigned char* pChar)
{
  return RxBuffer.GetChar (pChar);
}

void CUartDrv::ClearRxBuffer()
{
  RxBuffer.Clear();
}

void CUartDrv::ClearTxBuffer()
{
  TxBuffer.Clear();
}


int CUartDrv::OpenPort(const char* pDeviceName, const DCB* pOpenDcb)
{
  if ((!pDeviceName) || (!pOpenDcb))
    return 0;

  if (IsOpen)
    return 0;

// Open port

  CommHandle = CreateFile (pDeviceName, GENERIC_READ | GENERIC_WRITE, 0, NULL, 
                           OPEN_EXISTING, 
                           FILE_ATTRIBUTE_NORMAL | FILE_FLAG_OVERLAPPED,
                           NULL);

  if (CommHandle == INVALID_HANDLE_VALUE)
    return 0;

// Store original timeouts

  GetCommTimeouts (CommHandle, &OriginalTimeouts);

// Fill DCB

  memset (&dcb, 0, sizeof(dcb));
  dcb.DCBlength = sizeof(dcb);

  // Standard things: baud rate, character size, stop bits and parity
  dcb.fBinary       = TRUE;
  dcb.BaudRate      = pOpenDcb->BaudRate;
  dcb.ByteSize      = pOpenDcb->ByteSize;
  dcb.StopBits      = pOpenDcb->StopBits;
  dcb.fParity       = pOpenDcb->fParity;
  dcb.Parity        = pOpenDcb->Parity;

  // Flow control:
  // none:     RTS=DTR=ENABLE, I/O uncontrolled
  // XON/XOFF: RTS=DTR=ENABLE, I/O controlled by XON=0x11 and XOFF=0x13 characters
  // RTS/CTS:  DTR=ENABLE,     I/O controlled by RTS/CTS handshaking
  dcb.fOutxCtsFlow  = pOpenDcb->fOutxCtsFlow;
  if ((dcb.fRtsControl = pOpenDcb->fRtsControl) != RTS_CONTROL_HANDSHAKE)
    dcb.fRtsControl = RTS_CONTROL_ENABLE;
  dcb.fDtrControl   = DTR_CONTROL_ENABLE;
  dcb.fInX          = pOpenDcb->fInX;
  dcb.fOutX         = pOpenDcb->fOutX;
  dcb.XonChar       = 0x11;
  dcb.XoffChar      = 0x13;

// Set port state

  SetCommState (CommHandle, &dcb);

// Set new timeouts

  NewTimeouts.ReadIntervalTimeout = 0x01;
  NewTimeouts.ReadTotalTimeoutConstant =
    NewTimeouts.ReadTotalTimeoutMultiplier =
      NewTimeouts.WriteTotalTimeoutConstant =
        NewTimeouts.WriteTotalTimeoutMultiplier = 0;
  SetCommTimeouts (CommHandle, &NewTimeouts);

// Set port buffer sizes: Rx, Tx

  SetupComm (CommHandle, RX_BUFFER_SIZE, TX_BUFFER_SIZE);

// Create Rx and Tx threads

  pRxThread = AfxBeginThread ((AFX_THREADPROC)RxThreadFxn, this);
  if (pRxThread != NULL)
    pRxThread->m_bAutoDelete = FALSE;

  pTxThread = AfxBeginThread ((AFX_THREADPROC)TxThreadFxn, this);
  if (pTxThread != NULL)
    pTxThread->m_bAutoDelete = FALSE;

// Port is now really open

  IsOpen = 1;

  return 1;
}

int CUartDrv::ClosePort()
{
  if (!IsOpen)
    return 0;

  IsOpen = 0;

// Wait for Rx and Tx threads to exit

  QuitEvent.SetEvent();

  if (pRxThread != NULL)
  {
    WaitForSingleObject (pRxThread->m_hThread, THREAD_EXIT_TIMEOUT);
    delete pRxThread;
//    pRxThread = NULL;
  }

  if (pTxThread != NULL)
  {
    WaitForSingleObject (pTxThread->m_hThread, THREAD_EXIT_TIMEOUT);
    delete pTxThread;
//    pTxThread = NULL;
  }

  QuitEvent.ResetEvent();

// Drop DTR

  EscapeCommFunction (CommHandle, CLRDTR);

// Set original timeouts

  SetCommTimeouts (CommHandle, &OriginalTimeouts);

// Purge reads/writes, input buffer and output buffer

  PurgeComm (CommHandle, PURGE_TXABORT | PURGE_TXCLEAR | PURGE_RXABORT | PURGE_RXCLEAR);

// Close port

  CloseHandle (CommHandle);

  CommHandle = INVALID_HANDLE_VALUE;

  return 1;
}

CEvent& CUartDrv::GetRxEvent()
{
  return RxBuffer.NotEmpty;
}

typedef struct {
  int IsExitRequested;
  size_t Count;
} tXxResult;

static tXxResult RxFxn (unsigned char* pBuf, size_t Count,
                        HANDLE CommHandle, HANDLE QuitEventHandle)
{
  HANDLE Handles[2];
  OVERLAPPED ORead = {0};
  DWORD Read = 0;
  DWORD Res;
  tXxResult RxRes = {0, 0};

  Handles[0] = QuitEventHandle;
  Handles[1] = ORead.hEvent = CreateEvent (NULL, TRUE, FALSE, NULL);

  SetLastError (ERROR_SUCCESS);

  if (!ReadFile (CommHandle, pBuf, Count, &Read, &ORead))
  {
    if (GetLastError() == ERROR_IO_PENDING)
    {
      Res = WaitForMultipleObjects (2, Handles, FALSE, INFINITE);

      switch (Res)
      {
        case WAIT_OBJECT_0:
          // Exit requested
          RxRes.IsExitRequested = 1;
        break;

        case WAIT_OBJECT_0+1:
          // Read completed
          SetLastError (ERROR_SUCCESS);
          if (!GetOverlappedResult (CommHandle, &ORead, &Read, FALSE))
          {
            if (GetLastError() == ERROR_OPERATION_ABORTED)
              ErrorDlg ("Read: operation aborted!");
            else
              ErrorDlg ("Read: GetOverlappedResult() error!");
          }
          else
          {
            RxRes.Count = Read;
          }
        break;
      }
    }
  }
  else // if (!ReadFile (CommHandle, pBuf, Count, &Read, &ORead))
  {
    RxRes.Count = Read;
  }

  CloseHandle (ORead.hEvent);

  return RxRes;
}

static DWORD RxThreadFxn (LPVOID lParam)
{
  CUartDrv* pUartDrv = (CUartDrv*)lParam;
  unsigned char aBuf[READ_PORTION_MAX_SIZE];
  tXxResult RxRes;
  int IsDone = 0;

  while (!IsDone)
  {
    RxRes = RxFxn (aBuf, sizeof(aBuf), pUartDrv->CommHandle, pUartDrv->QuitEvent);

    if (RxRes.IsExitRequested)
    {
      // Exit requested
      IsDone = 1;
    }
    else
    {
      // Read completed
      if (RxRes.Count)
      {
        unsigned i;
        for (i=0;i<RxRes.Count;i++)
          pUartDrv->RxBuffer.PutChar (aBuf[i]);
      }
    }
  }

  return 0;
}

static tXxResult TxFxn (unsigned char* pBuf, size_t Count,
                        HANDLE CommHandle, HANDLE QuitEventHandle)
{
  HANDLE Handles[2];
  OVERLAPPED OWrite = {0};
  DWORD Written = 0;
  DWORD Res;
  tXxResult TxRes = {0, 0};

  Handles[0] = QuitEventHandle;
  Handles[1] = OWrite.hEvent = CreateEvent (NULL, TRUE, FALSE, NULL);
  
  SetLastError (ERROR_SUCCESS);

  if (!WriteFile (CommHandle, pBuf, Count, &Written, &OWrite))
  {
    if (GetLastError() == ERROR_IO_PENDING)
    {
      Res = WaitForMultipleObjects (2, Handles, FALSE, INFINITE);
      switch (Res)
      {
        case WAIT_OBJECT_0:
          // Exit requested
          TxRes.IsExitRequested = 1;
        break;

        case WAIT_OBJECT_0+1:
          // Write completed
          SetLastError (ERROR_SUCCESS);
          if (!GetOverlappedResult (CommHandle, &OWrite, &Written, FALSE))
          {
            if (GetLastError() == ERROR_OPERATION_ABORTED)
              ErrorDlg ("Write: operation aborted!");
            else
              ErrorDlg ("Write: GetOverlappedResult() error!");
          }

          if (Written != Count)
          {
            if (GetLastError() == ERROR_SUCCESS)
              ErrorDlg ("Write: operation timed out!");
            else
              ErrorDlg ("Write: operation error!");
          }
        break;
      }
    }
    else // if (GetLastError() == ERROR_IO_PENDING)
    {
      ErrorDlg ("Write: GetLastError() != ERROR_IO_PENDING!");
    }
  }

  CloseHandle (OWrite.hEvent);

  TxRes.Count = Written;

  return TxRes;
}

static DWORD TxThreadFxn (LPVOID lParam)
{
  CUartDrv* pUartDrv = (CUartDrv*)lParam;
  HANDLE Handles[2] = { pUartDrv->QuitEvent, pUartDrv->TxBuffer.NotEmpty };
  DWORD Res;
  unsigned char c = 0;
  tXxResult TxRes;
  int IsDone = 0;
  
  while (!IsDone)
  {
    Res = WaitForMultipleObjects (2, Handles, FALSE, INFINITE);
    switch (Res)
    {
      case WAIT_OBJECT_0:
        // Exit requested
        IsDone = 1;
      break;

      case WAIT_OBJECT_0+1:
        // Tx requested
        pUartDrv->TxBuffer.GetChar (&c);
        TxRes = TxFxn (&c, sizeof(c), pUartDrv->CommHandle, Handles[0]);
        if (TxRes.IsExitRequested)
          IsDone = 1;
      break;
    }
  }

  return 0;
}
